require('dotenv').config();
require('module-alias/register');
const { log } = require('@lib/logger');
const EventEmitter = require('node:events')
const { checkEnv } = require('dotenv-joi');
const envSchema = require('./envSchema');
const os = require('node:os')
const randomstring = require('randomstring')

process.stdin.resume();

// Init logger
process.log = log;

const { getChampionData } = require('@lib/card_utils');
const { prepareRoot } = require('@lib/proto');

const randomSlug = randomstring.generate({
  length: 8,
  charset: 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'
});

let userSlug = null;
// Check if nodeName is larger than 8 and only contains abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890
if(process.env.NODENAME?.length > 8 && process.env.NODENAME.match(/^[a-zA-Z0-9]+$/)) userSlug = process.env.NODENAME;


const myName = `node.${os.hostname}.${userSlug || randomSlug}`
process.me = myName

/*
if (!fs.existsSync('.me')) {
  const randomSlug = randomstring.generate({
    length: 8,
    charset: 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'
});
  const myName = `node.${os.hostname}.${randomSlug}`
  fs.writeFileSync(path.join(__dirname, '.me'), myName)
  process.me = myName
} else {
  const myName = fs.readFileSync(path.join(__dirname, '.me'), 'utf-8')
  process.me = myName
}
*/

const isENVok = checkEnv(envSchema);
if (isENVok) {
  process.log.error('Invalid ENV values. Please check your .env file and fix the above mentioned errors.');
  process.exit(1);
}

// Load more complex modules
//const watchdog = require('@lib/watchdog');

// Prepare process.app
process.app = {};
//process.app.watchdog = watchdog;

process.eventbus = new EventEmitter;

// WINDOWS ONLY
if (process.platform === "win32") {
  var rl = require("readline").createInterface({
    input: process.stdin,
    output: process.stdout
  });

  rl.on("SIGINT", function () {
    process.emit("SIGINT");
  });
}

// Go application livecycle

const WebSocketManager = require('@src/communication');
const { PuppeteerClusterManager } = require('@src/puppeteer');

/* BENCMARKING */

/**
 * Benchmark Funktion
 * @param {Number} concurrency 
 * @param {Number} jobCount 
 * @returns 
 */
const benchmarkConcurrency = async (concurrency, jobCount) => {
  const puppeteer = new PuppeteerClusterManager(concurrency);
  await puppeteer.init(true);

  const start_benchmark = Date.now();
  const promises = [];
  for (let i = 0; i < jobCount; i++) {
    const promise = puppeteer.addQueue(i,
      [
        {
          uuid: 'D1BN-E0X5ZJxwFjETUa1g9mh4H9Y0yx_A7YIEN7lf8M8wcX3rXgrY37pGVgPT93oEZG_SVZkk_fR7w',
          shard: 'euw'
        }
      ], "de-DE", "euw", "dark")
    promises.push(promise);
  }

  await Promise.all(promises);

  await puppeteer.idle();
  const benchmark_time = (Date.now() - start_benchmark) / 1000;
  await puppeteer.end();

  return benchmark_time;
}

/**
 * Benchmark the current system to find the optimal concurrency.
 * @returns {number} - The optimal concurrency for the current system.
 */
const findOptimalConcurrency = async () => {
  const jobIncrements = [1, 80, 100, 120];
  let bestConcurrency = 2;
  let previousTime = Infinity;

  for (const jobCount of jobIncrements) {
    let increasing = false;
    for (let concurrency = 2; concurrency <= 14; concurrency++) {
      const currentTime = await benchmarkConcurrency(concurrency, jobCount);
      process.log.system(`Concurrency: ${concurrency}, Jobs: ${jobCount}, Time: ${currentTime}s, Performance: ${jobCount / currentTime}`);

      if (currentTime < previousTime) {
        bestConcurrency = concurrency;
        previousTime = currentTime;
      } else {
        increasing = true;
        break; // Early break, since we know that the time is increasing
      }
    }

    // Don´t even test higher job counts if we got a clear winner
    if (increasing) {
      break;
    }
  }

  return bestConcurrency;
}

/* LIVECYCLE */
(async () => {
  try {
    // If the loadbalancer isn´t running after 15 seconds, we assume that it´s down and exit.
    const firstConnectFailed = setTimeout(() => {
      process.log.error('Unknown state. Loadbalancer didn´t respond in time. Shutting down...');
      process.exit(404);
    }, 15 * 1000);

    const com = new WebSocketManager(process.env.WS_PROTOCOL, process.env.BALANCER_IP, process.env.BALANCER_PORT);
    await com.connect();

    // Reset the timeout (Because we got a connection)
    clearTimeout(firstConnectFailed);
    // Download champion.json from CDN
    await getChampionData('en_US');
    // Start Benchmarking
    com.setState('benchmarking');
    await prepareRoot(); // Prepare the root belt for the proto files
    const best_concurrency = await findOptimalConcurrency();
    process.log.system(`Best concurrency: ${best_concurrency}`);
    com.setState('starting');
    // Start Puppeteer
    const puppeteer = new PuppeteerClusterManager(best_concurrency);
    await puppeteer.init();

    com.setPuppeteer(puppeteer);

    // Graceful shutdown
    process.on('SIGTERM', async () => {
      process.log.system('SIGTERM received. Shutting down...');
      com.setState('shutting_down');
      process.log.system('Waiting for all jobs to finish...');
      await puppeteer.end(true);
      process.exit();
    });

    process.on('SIGINT', async () => {
      process.log.system('SIGINT received. Shutting down...');
      com.setState('shutting_down');
      process.log.system('Waiting for all jobs to finish...');
      await puppeteer.end(true);
      process.log.system('All jobs finished. Shutting down...');
      process.exit();
    });

  } catch (err) {
    process.log.error(err);
    process.exit(1);
  }
})();